# encoding: utf-8
# Copyright (C) 2015 John Törnblom
'''
Serialize xtuml models to an sql-based file format and persist to disk.
'''


import uuid
import logging


logger = logging.getLogger(__name__)


def serialize_value(value):
    if   isinstance(value, bool):
        return '%d' % int(value)

    elif isinstance(value, str):
        return "'%s'" % value.replace("'", "''")

    elif isinstance(value, int):
        return '%d' % value
    
    elif isinstance(value, float):
        return '%f' % value

    elif isinstance(value, uuid.UUID):
        return '"%s"' % value


def serialize_instance(inst, table, attributes):
    attr_count = 0
    attributes = list(attributes)

    s = 'INSERT INTO %s VALUES (' % table
    for name, ty in attributes:
        value = getattr(inst, name)
        s += '\n    '
        s += serialize_value(value)

        attr_count += 1
        if attr_count < len(attributes):
            s += ', -- %s : %s' % (name, ty)
        else:
            s += ' -- %s : %s' % (name, ty)

    s += '\n);\n'

    return s


def serialize_metamodel(metamodel):
    s = ''
    for lst in metamodel.instances.values():
        for inst in lst:
            table = inst.__class__.__name__
            params = metamodel.param_names[table]
            types = metamodel.param_types[table]
            s += serialize_instance(inst, table, zip(params, types))
            
    return s


def persist_metamodel(metamodel, path):
    with open(path, 'w') as f:
        s = serialize_metamodel(metamodel)
        f.write(s)




# encoding: utf-8
# Copyright (C) 2015 John Törnblom

class Node(object):
    '''
    Tree node which may be walked a walker, and visited by a visitor.
    '''
    
    def __str__(self):
        return self.__class__.__name__


class Visitor(object):
    '''
    A visitor may be used to visit tree nodes walked by a walker.
    '''
    
    def enter(self, node):
        name = 'enter_' + node.__class__.__name__
        fn = getattr(self, name, self.default_enter)
        fn(node)

    def leave(self, node):
        name = 'leave_' + node.__class__.__name__
        fn = getattr(self, name, self.default_leave)
        fn(node)

    def default_enter(self, node):
        pass

    def default_leave(self, node):
        pass
    
    
class Walker(object):
    '''
    A walker may be used to walk a tree.
    '''
    
    def __init__(self):
        self.visitors = list()
        
    def accept(self, node, **kwargs):
        if node is None:
            return
        
        for v in self.visitors: v.enter(node)
        
        name = 'accept_' + node.__class__.__name__
        fn = getattr(self, name, self.default_accept)
        r = fn(node, **kwargs)
        
        for v in self.visitors: v.leave(node)
        
        return r
    
    def default_accept(self, node, **kwargs):
        for child in filter(None, node.children):
            self.accept(child, **kwargs)


class NodePrintVisitor(Visitor):
    '''
    A visitor that prints a tree-like structure to stdout.
    '''
    
    def __init__(self):
        self.__lvl = 0
        
    def default_enter(self, node):
        print('%s%s' % ("  " * self.__lvl, node))
        self.__lvl += 1

    def default_leave(self, node):
        self.__lvl -= 1
    



# encoding: utf-8
# Copyright (C) 2015 John Törnblom

from .load import load_metamodel
from .load import ParsingException
from .persist import persist_metamodel
from .persist import serialize_metamodel
from .persist import serialize_instance
from .persist import serialize_value

from .model import MetaModel
from .model import IdGenerator
from .model import ModelException
from .model import navigate_any
from .model import navigate_one
from .model import navigate_many

from . import version



# xtuml.__xtuml_lextab.py. This file automatically created by PLY (version 3.4). Don't edit!
_tabversion   = '3.2'
_lextokens    = {'NUMBER': 1, 'PHRASE': 1, 'GUID': 1, 'MINUS': 1, 'INSERT': 1, 'RPAREN': 1, 'SEMICOLON': 1, 'CREATE': 1, 'REF_ID': 1, 'TO': 1, 'COMMA': 1, 'FRACTION': 1, 'ROP': 1, 'TABLE': 1, 'STRING': 1, 'INTO': 1, 'LPAREN': 1, 'ID': 1, 'RELID': 1, 'FROM': 1, 'VALUES': 1, 'CARDINALITY': 1}
_lexreflags   = 0
_lexliterals  = ''
_lexstateinfo = {'INITIAL': 'inclusive'}
_lexstatere   = {'INITIAL': [('(?P<t_comment>\\-\\-([^\\n]*\\n))|(?P<t_COMMA>,)|(?P<t_FRACTION>(\\d+)(\\.\\d+))|(?P<t_RELID>R[0-9]+)|(?P<t_CARDINALITY>(1C))|(?P<t_ID>[A-Za-z_][\\w_]*)|(?P<t_LPAREN>\\()|(?P<t_MINUS>-)|(?P<t_NUMBER>[0-9]+)|(?P<t_RPAREN>\\))|(?P<t_SEMICOLON>;)|(?P<t_STRING>\\\'((\\\'\\\')|[^\\\'])*\\\')|(?P<t_GUID>\\"([^\\\\\\n]|(\\\\.))*?\\")|(?P<t_newline>\\n+)', [None, ('t_comment', 'comment'), None, ('t_COMMA', 'COMMA'), ('t_FRACTION', 'FRACTION'), None, None, ('t_RELID', 'RELID'), ('t_CARDINALITY', 'CARDINALITY'), None, ('t_ID', 'ID'), ('t_LPAREN', 'LPAREN'), ('t_MINUS', 'MINUS'), ('t_NUMBER', 'NUMBER'), ('t_RPAREN', 'RPAREN'), ('t_SEMICOLON', 'SEMICOLON'), ('t_STRING', 'STRING'), None, None, ('t_GUID', 'GUID'), None, None, ('t_newline', 'newline')])]}
_lexstateignore = {'INITIAL': ' \t\r\x0c'}
_lexstateerrorf = {'INITIAL': 't_error'}



#!/usr/bin/env python
# encoding: utf-8
# Copyright (C) 2015 John Törnblom
import logging
import unittest
import sys

from distutils.core import setup
from distutils.core import Command

import xtuml


logging.basicConfig(level=logging.DEBUG)


class PrepareCommand(Command):
    description = "Prepare the source code by generating lexers and parsers"
    user_options = []

    def initialize_options(self):
        pass
    
    def finalize_options(self):
        pass

    def run(self):
        xtuml.load_metamodel([])


class TestCommand(Command):
    description = "Execute unit tests"
    user_options = []

    def initialize_options(self):
        pass
    
    def finalize_options(self):
        pass

    def run(self):
        suite = unittest.TestLoader().discover('tests')
        runner = unittest.TextTestRunner(verbosity=2, buffer=True)
        exit_code = not runner.run(suite).wasSuccessful()
        sys.exit(exit_code)


setup(name='pyxtuml',
      version=xtuml.version.release,
      description='Library for parsing, manipulating, and generating BridgePoint xtUML models',
      author='John Törnblom',
      author_email='john.tornblom@gmail.com',
      url='https://github.com/john-tornblom/pyxtuml',
      license='GPLv3',
      classifiers=[
          'Development Status :: 4 - Beta',
          'Intended Audience :: Developers',
          'Topic :: Software Development :: Code Generators',
          'Topic :: Software Development :: Compilers',
          'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
          'Programming Language :: Python :: 2.7',
          'Programming Language :: Python :: 3.4'],
      keywords='xtuml bridgepoint',
      platforms=["Linux"],
      packages=['xtuml'],
      requires=['ply'],
      cmdclass={'prepare': PrepareCommand, 'test': TestCommand}
      )




# encoding: utf-8
# Copyright (C) 2015 John Törnblom

from itertools import chain

import logging
import uuid
import collections

try:
    from future_builtins import filter, zip
except ImportError:
    pass


logger = logging.getLogger(__name__)


class ModelException(Exception):
    pass


class NavChain(object):
    
    def __init__(self, handle, is_many=True):
        if handle is None:
            self.handle = QuerySet()
        
        elif isinstance(handle, BaseObject):
            self.handle = QuerySet([handle])
        
        elif isinstance(handle, QuerySet):
            self.handle = handle
        
        else:
            raise ModelException("Unable to navigate instances of '%s'" % type(handle))
            
        self.is_many = is_many
        self._kind = None
    
    def nav(self, kind, relid, phrase=''):
        if isinstance(relid, int):
            relid = 'R%d' % relid
        
        result = QuerySet()
        for child in iter(self.handle):
            result |= child.__q__[kind][relid][phrase](child)

        self.handle = result
        
        return self
    
    def __getattr__(self, name):
        self._kind = name
        return self
    
    def __getitem__(self, args):
        if not isinstance(args, tuple):
            args = (args, '')
        
        relid, phrase = args
        
        return self.nav(self._kind, relid, phrase)

    def __call__(self, where_clause=None):
        s = filter(where_clause, self.handle)
        if self.is_many:
            return QuerySet(s)
        else:
            return next(s, None)


def navigate_one(inst):
    return navigate_any(inst)


def navigate_any(inst):
    return NavChain(inst, is_many=False)


def navigate_many(inst):
    return NavChain(inst, is_many=True)


class AssociationEndPoint(object):
    
    def __init__(self, kind, cardinality, ids, phrase=''):
        self.cardinality = cardinality
        self.kind = kind
        self.phrase = phrase
        self.ids = ids

    @property
    def is_many(self):
        return self.cardinality.upper() in ['M', 'MC']

    @property
    def is_conditional(self):
        return 'C' in self.cardinality.upper()


class SingleEndPoint(AssociationEndPoint):
    
    def __init__(self, kind, conditional=False, ids=[], phrase=''):
        if conditional: cardinality = '1C'
        else:           cardinality = '1'
        AssociationEndPoint.__init__(self, kind, cardinality, ids, phrase)


class ManyEndPoint(AssociationEndPoint):
    
    def __init__(self, kind, conditional=False, ids=[], phrase=''):
        if conditional: cardinality = 'MC'
        else:           cardinality = 'M'
        AssociationEndPoint.__init__(self, kind, cardinality, ids, phrase)
        

class OrderedSet(collections.MutableSet):
    # Originally posted on http://code.activestate.com/recipes/576694
    # by Raymond Hettinger.
    
    def __init__(self, iterable=None):
        self.end = end = [] 
        end += [None, end, end]         # sentinel node for doubly linked list
        self.map = {}                   # key --> [key, prev, next]
        if iterable is not None:
            self |= iterable

    def __len__(self):
        return len(self.map)

    def __contains__(self, key):
        return key in self.map

    def add(self, key):
        if key not in self.map:
            end = self.end
            curr = end[1]
            curr[2] = end[1] = self.map[key] = [key, curr, end]

    def discard(self, key):
        if key in self.map:        
            key, prev, next_ = self.map.pop(key)
            prev[2] = next_
            next[1] = prev

    def __iter__(self):
        end = self.end
        curr = end[2]
        while curr is not end:
            yield curr[0]
            curr = curr[2]

    def __reversed__(self):
        end = self.end
        curr = end[1]
        while curr is not end:
            yield curr[0]
            curr = curr[1]

    def pop(self, last=True):
        if not self:
            raise KeyError('set is empty')
        
        if last:
            key = self.end[1][0]
        else:
            key = self.end[2][0]
            
        self.discard(key)
        return key

    def __repr__(self):
        if not self:
            return '%s()' % (self.__class__.__name__,)
        return '%s(%r)' % (self.__class__.__name__, list(self))

    def __eq__(self, other):
        if isinstance(other, OrderedSet):
            return len(self) == len(other) and list(self) == list(other)
        return set(self) == set(other)


class QuerySet(OrderedSet):
    
    @property
    def first(self):
        if len(self):
            return next(iter(self))
    
    @property
    def last(self):
        if len(self):
            return next(reversed(self))


class BaseObject(object):
    __q__ = None
    __c__ = None
    
    def __init__(self):
        self.__c__.clear()
        
    def __add__(self, other):
        assert isinstance(other, BaseObject)
        return QuerySet([self, other])

    def __sub__(self, other):
        assert isinstance(other, BaseObject)
        if self == other: return QuerySet()
        else: return QuerySet([self])

    def __getattr__(self, name):
        lname = name.lower()
        for attr in self.__dict__.keys():
            if attr.lower() == lname:
                return self.__dict__[attr]

        object.__getattribute__(self, name)
    
    def __setattr__(self, name, value):
        name = name.lower()
        for attr in self.__dict__.keys():
            if attr.lower() == name:
                self.__dict__[attr] = value
                self.__c__.clear()
                return

    def __str__(self):
        return str(self.__dict__)
    
    
class IdGenerator(object):
    
    def __init__(self, readfunc=uuid.uuid4):
        self.readfunc = readfunc
        self._current = readfunc()
    
    def peek(self):
        return self._current
    
    def next(self):
        val = self._current
        self._current = self.readfunc()
        return val
    
    def __iter__(self):
        return self

    def __next__(self):
        return self.next()


class MetaModel(object):
    
    classes = None
    instances = None
    param_names = None
    param_types = None
    id_generator = None
    
    def __init__(self, id_generator=IdGenerator()):
        self.classes = dict()
        self.instances = dict()
        self.param_names = dict()
        self.param_types = dict()
        self.id_generator = id_generator
        
    def define_class(self, kind, attributes):
        Cls = type(kind, (BaseObject,), dict(__q__=dict(), __c__=dict()))
        self.classes[kind] = Cls
        self.param_names[kind] = [name for name, _ in attributes]
        self.param_types[kind] = [ty for _, ty in attributes]
        
        return Cls
    
    def default_value(self, ty_name):
        if   ty_name == 'boolean': return False
        elif ty_name == 'integer': return 0
        elif ty_name == 'integer': return 0
        elif ty_name == 'real': return 0.0
        elif ty_name == 'string': return ''
        elif ty_name == 'unique_id': return next(self.id_generator)
        else: raise ModelException("Unknown type named '%s'" % ty_name)
    
    def type_name(self, ty):
        if   issubclass(ty, bool): return 'boolean'
        elif issubclass(ty, int): return 'integer'
        elif issubclass(ty, float): return 'real'
        elif issubclass(ty, str): return 'string'
        elif issubclass(ty, BaseObject): return 'inst_ref'
        elif issubclass(ty, type(None)): return 'inst_ref'
        elif issubclass(ty, QuerySet): return 'inst_ref_set'
        elif issubclass(ty, type(self.id_generator.peek())): return 'unique_id'
        else: raise ModelException("Unknown type '%s'" % ty.__name__)
        
    def named_type(self, name):
        lookup_table = {
          'boolean'     : bool,
          'integer'     : int,
          'real'        : float,
          'string'      : str,
          'inst_ref'    : BaseObject,
          'inst_ref_set': QuerySet,
          'unique_id'   : type(self.id_generator.peek())
        }
        
        if name in lookup_table:
            return lookup_table[name]
        else:
            raise ModelException("Unknown type named '%s'" % name)
        
    def new(self, kind, *args, **kwargs):
        if kind not in self.classes:
            raise ModelException("Unknown class %s" % kind)
        
        Cls = self.classes[kind]
        inst = Cls()
        
        # set all parameters with an initial default value
        for key, ty in zip(self.param_names[kind], self.param_types[kind]):
            inst.__dict__[key] = self.default_value(ty)

        # set all positional arguments
        for name, ty, value in zip(self.param_names[kind], self.param_types[kind], args):
            Type = self.named_type(ty)
            if not isinstance(value, Type):
                value = Type(value)
                
            inst.__dict__[name] = value

        # set all named arguments
        inst.__dict__.update(kwargs)

        if not kind in self.instances:
            self.instances[kind] = list()
            
        self.instances[kind].append(inst)
        
        return inst
        
    @staticmethod
    def empty(inst):
        if   inst is None: return True
        elif isinstance(inst, QuerySet): return len(inst) == 0
        return False
    
    @staticmethod
    def not_empty(inst):
        return not MetaModel.empty(inst)
    
    @staticmethod
    def cardinality(inst):
        if   inst is None: return 0
        elif isinstance(inst, QuerySet): return len(inst)
        else: return 1

    @staticmethod
    def is_set(inst):
        return isinstance(inst, QuerySet)
    
    @staticmethod
    def is_instance(inst):
        return isinstance(inst, BaseObject)
    
    @staticmethod
    def first(inst, query_set):
        assert isinstance(query_set, QuerySet)
        return inst == query_set.first
    
    @staticmethod
    def not_first(inst, query_set):
        assert isinstance(query_set, QuerySet)
        return inst != query_set.first
    
    @staticmethod
    def last(inst, query_set):
        assert isinstance(query_set, QuerySet)
        return inst == query_set.last

    @staticmethod
    def not_last(inst, query_set):
        assert isinstance(query_set, QuerySet)
        return inst != query_set.last
    
    def _query(self, kind, many, **kwargs):
        for inst in iter(self.instances[kind]):
            for name, value in kwargs.items():
                if getattr(inst, name) != value:
                    break
            else:
                yield inst
                if not many:
                    return
                    
    def _select_endpoint(self, inst, source, target, kwargs):
        if not target.kind in self.instances:
            return QuerySet()
        
        keys = chain(target.ids, kwargs.keys())
        values = chain([getattr(inst, name) for name in source.ids], kwargs.values())
        kwargs = dict(zip(keys, values))
                            
        cache_key = frozenset(list(kwargs.items()))
        cache = self.classes[target.kind].__c__
        
        if cache_key not in cache:
            cache[cache_key] = QuerySet(self._query(target.kind, target.is_many, **kwargs))
            
        return cache[cache_key]
    
    def _formalized_query(self, source, target):
        return lambda inst, **kwargs: self._select_endpoint(inst, source, target, kwargs)
    
    def define_relation(self, rel_id, end1, end2):
        Cls1 = self.classes[end1.kind]
        Cls2 = self.classes[end2.kind]

        if end2.kind not in Cls1.__q__:
            Cls1.__q__[end2.kind] = dict()
            
        if end1.kind not in Cls2.__q__:
            Cls2.__q__[end1.kind] = dict()
        
        if rel_id not in Cls1.__q__[end2.kind]:
            Cls1.__q__[end2.kind][rel_id] = dict()
            
        if rel_id not in Cls2.__q__[end1.kind]:
            Cls2.__q__[end1.kind][rel_id] = dict()
        
        Cls1.__q__[end2.kind][rel_id][end2.phrase] = self._formalized_query(end1, end2)
        Cls2.__q__[end1.kind][rel_id][end1.phrase] = self._formalized_query(end2, end1)
    
    def select_one(self, kind, where_cond=None):
        return self.select_any(kind, where_cond)
    
    def select_any(self, kind, where_cond=None):
        if kind in self.instances:
            s = filter(where_cond, self.instances[kind])
            return next(s, None)
        
    def select_many(self, kind, where_cond=None):
        if kind in self.instances:
            return QuerySet(filter(where_cond, self.instances[kind]))
        else:
            return QuerySet()


